--//=======================================================
--//  SOME USEFULL LIBRARY DECLARATIONS (YOU MAY NOT NEED 
--//  ALL OF THEM IN YOUR DESIGN)
--//=======================================================
LIBRARY ieee;
    USE ieee.std_logic_1164.ALL;
    USE ieee.std_logic_unsigned.all ; 
    USE ieee.std_logic_arith.all ;
    use IEEE.std_logic_textio.all;

LIBRARY std;
    USE std.textio.all;
	 
library altera_mf;
use altera_mf.altera_mf_components.all;


ENTITY serial IS
   
--//=======================================================
--//  PORT declarations
--//=======================================================

   PORT (

--	//////////// CLOCK //////////
	CLOCK_50 		: IN STD_LOGIC;
--	//////////// LED //////////
	LED			:OUT STD_LOGIC_VECTOR(7 downto 0);
--	//////////// GPIO_0, GPIO_0 connect to GPIO Default //////////
	--GPIO_0			:INOUT STD_LOGIC_VECTOR(33 downto 0);
	GPIO_0			:IN STD_LOGIC_VECTOR(33 downto 0);	-- SIMULATION ONLY
	GPIO_0_IN 		:IN STD_LOGIC_VECTOR(1 downto 0);
--	//////////// GPIO_1, GPIO_1 connect to GPIO Default //////////
	--GPIO_1			:INOUT STD_LOGIC_VECTOR(33 downto 0);
	GPIO_1			:IN STD_LOGIC_VECTOR(33 downto 0);	-- SIMULATION ONLY
	GPIO_1_IN		:IN STD_LOGIC_VECTOR(1 downto 0)
);

end serial;

ARCHITECTURE arch OF serial IS

--//=======================================================
--//  COMPONENT declarations
--//=======================================================

--//=======================================================
--// SIGNAL declarations
--//=======================================================
signal led_s : STD_LOGIC_VECTOR(7 downto 0);
signal UserIF : STD_LOGIC_VECTOR(3 downto 0);
signal DataIF : STD_LOGIC_VECTOR(3 downto 0);
signal cnt : STD_LOGIC_VECTOR(6 downto 0);
signal clki : STD_LOGIC ;
signal rst : STD_LOGIC ;
----- Serial i/f
signal eRDDreq : STD_LOGIC ;
signal eRDDval : STD_LOGIC ;
signal eRDData : STD_LOGIC;
signal restart : STD_LOGIC;
----- RD user control i/f
signal eRDQWen : STD_LOGIC ; -- Read Quad-Word (4 Bytes) from serial I/F 
signal iRDWreq : STD_LOGIC ; -- Read word (1 Byte) from Memory and output to LED
signal iRDaddr : STD_LOGIC_VECTOR(1 downto 0) ; -- Memory Addr to read

-- Internal signals
signal eRDQWen_d, eRDDval_d: STD_LOGIC ;
signal eRDQWen_2d, eRDDval_2d: STD_LOGIC ;
signal New_RD_flag, New_Data_flag: STD_LOGIC ;
signal data_reg, DQWord: STD_LOGIC_VECTOR(31 downto 0) ;
signal test: STD_LOGIC ;

-- FSM signals 

TYPE Gstate_type IS (IDLE,NEW_RDQW, GW8);
signal extIF_Gstate_cur, extIF_Gstate_nxt	   : Gstate_type;
TYPE Rstate_type IS (IDLE,START,W8,DATA);
signal extIF_Rstate_cur, extIF_Rstate_nxt	   : Rstate_type;



	
BEGIN

clki <= CLOCK_50;
UserIF <= GPIO_0(3 downto 0);
DataIF <= GPIO_0(7 downto 4);
rst <= GPIO_0(13);

eRDQWen<=UserIF(0);		-- User request to initiate new word trasfer
iRDWreq<=UserIF(1);		-- User request to read word from internal memory
iRDaddr<=UserIF(3 downto 2); 	-- Memory Addr to read
--DataIF(0)<=eRDDreq;
eRDDval<=DataIF(1);		-- Dvalid
eRDData<=DataIF(2);		-- Data
--DataIF(3)<=restart ;


-- =============== MOTHER PROCESS ================
extRD_TOPproc: process (extIF_Gstate_cur, extIF_Rstate_cur, New_RD_flag)
begin
	case extIF_Gstate_cur is
	when IDLE => -- Idle
		if (New_RD_flag='1') then
	  	   extIF_Gstate_nxt <= NEW_RDQW; --New activation of eRDWen
	 	else
		   extIF_Gstate_nxt <= IDLE; 
		end if;
	when NEW_RDQW => -- uRDWen
		if (extIF_Rstate_cur=START) then
	  	   extIF_Gstate_nxt <= GW8;
	 	else
		   extIF_Gstate_nxt <= NEW_RDQW;
		end if;
	when GW8 => -- Serial word RD in progress
		if (extIF_Rstate_cur=IDLE) then
	  	   extIF_Gstate_nxt <= IDLE; 	-- Transaction ended
	 	else
		   extIF_Gstate_nxt <= GW8; 
		end if;
	end case;
end process extRD_TOPproc;

-- =============== CHILD PROCESS (MAIN QW RD, Serial protocol execution FSM) ================
extRD_MAINproc: process (extIF_Gstate_cur, extIF_Rstate_cur, New_Data_flag, cnt)
begin
	case extIF_Rstate_cur is
	when IDLE => -- Idle
		data_reg <= (others=>'0'); 	-- Clear data_reg
		--if (cnt=8) then DQWord <= data_reg; end if;
		if (extIF_Gstate_cur=NEW_RDQW) then	--New activation of eRDWen
	  	   extIF_Rstate_nxt <= START; 
	 	else
		   extIF_Rstate_nxt <= IDLE; --Idle
		end if;
	when START => 
		extIF_Rstate_nxt <= W8; 
	when W8 => 
		if (New_Data_flag='1') then	
	  	   extIF_Rstate_nxt <= DATA; 
	 	else
		   extIF_Rstate_nxt <= W8;
		end if;
	when DATA => 
		data_reg <= data_reg(30 downto 0) & eRDData;	-- Shift and append new bit
		if (cnt=31) then	
		   extIF_Rstate_nxt <= IDLE;
	 	else
		   extIF_Rstate_nxt <= W8; 
		end if;
	end case;

end process extRD_MAINproc;

-- =============== MOORE FSM ADVANCE STATE PROCESS ================
extRDFSMctrl: process(clki, rst)
begin
   if (rst='1') then
      extIF_Gstate_cur <= IDLE;
      extIF_Rstate_cur <= IDLE;
   elsif (clki'event and clki = '1') then
	extIF_Gstate_cur <= extIF_Gstate_nxt; 
   	extIF_Rstate_cur <= extIF_Rstate_nxt;
   end if;
end process extRDFSMctrl;


-- =============== SIGNAL TRANSITION MONITORING PROCESS ================
flaggen: process(clki, rst)
begin
if (rst='1') then
   New_RD_flag <= '0';
   eRDQWen_d <= '0';  -- 1 clk delay
   New_Data_flag <= '0';
   eRDDval_d <= '0';  -- 1 clk delay
   eRDQWen_2d <= '0';  -- 1 clk delay
   eRDDval_2d <= '0';  -- 1 clk delay
   cnt <= (others=>'0'); 		-- Restart counting
   restart <= '0';
   test <= '0';  -- 1 clk delay
   DQWord <= (others=>'0');
elsif (clki'event and clki = '1') then
	eRDQWen_d  <= eRDQWen ;  -- 1 clk delay
   	eRDDval_d  <= eRDDval ;  -- 1 clk delay
	eRDQWen_2d  <= eRDQWen_d ;  -- 2 clk delay
   	eRDDval_2d  <= eRDDval_d ;  -- 2 clk delay
	if (eRDQWen='1') AND (eRDQWen_2d='0') then	-- Activates for 2 clk after eRDQWen rising edge
		New_RD_flag <='1';
		test <= '1';
	else 
		New_RD_flag <='0';
		test <= test;
	end if;
	if (eRDDval='1') AND (eRDDval_2d='0') then	-- Activates for 2 clk after eRDDval rising edge
		New_Data_flag <='1';
	else 
		New_Data_flag <='0';
	end if;
	if (extIF_Rstate_cur=IDLE) OR (extIF_Rstate_cur=START) then
		restart <= '1';
	elsif (extIF_Rstate_cur=W8) then
		if (cnt="0000000") then 
			restart <= '1';
		else
			restart <= '0';
		end if;
	else 
		restart <= '0';
	end if;

	case extIF_Rstate_cur is
	when IDLE => -- Idle
		cnt <= (others=>'0'); 		-- Restart counting
	when START => 
		cnt <= (others=>'0'); 		-- Restart counting
	when W8 => 
		cnt <= cnt; 	
	when DATA => 
		cnt <=cnt+1;	
		if (cnt=31) then DQWord <= data_reg; end if;
	end case;
end if;
end process flaggen;


-- =============== OUTPUT SELECT PROCESS ================
outp: process(iRDWreq, iRDWreq, extIF_Rstate_cur)
begin
   case extIF_Rstate_cur is 
      when W8 =>
	led_s <= "0000000" & eRDDreq;
      when others => 
	if (iRDWreq='1') then
	   case iRDaddr is
	   when "00" => 
		led_s <= DQWord(7 downto 0);
	   when "01" => 
		led_s <= DQWord(15 downto 8);
	   when "10" => 
		led_s <= DQWord(23 downto 16);
	   when others => 
		led_s <= DQWord(31 downto 24);
	   end case;
	else
	   led_s <= (others=>'0');
	end if;
   end case;

end process ;

with extIF_Rstate_cur select eRDDreq <= '0' when IDLE, '1' when others; 
LED <= led_s;


END arch;
